<?php
namespace Teknisa\MAN\Controllers;

use Teknisa\MAN\Services\TransmissaoNfceService;
use Zeedhi\Framework\DTO\Request;
use Zeedhi\Framework\DTO\Response;
use Zeedhi\Framework\DataSource\DataSet;
use Zeedhi\Framework\DTO\Response\Notification;
use Teknisa\MAN\Utilities\Resposta;
use Teknisa\MAN\Utilities;
use Teknisa\MAN\Factories\TransmissaoNFCE;
use Teknisa\MAN\Factories\CorrigirNFCE;
use Teknisa\MAN\Factories\NotaFiscal;
use Zeedhi\Framework\DTO\Response\Message;

class TransmissaoNfceController {
    
    private $transmissaoService, $utilService, $entityManager, $notaFiscal;
    
    private $defaultFields = array(
        "SITUACAO"        => "P_R_T_A_D_C",
        "CDFILIAL"        => "T",
        "CDCAIXA"         => "T",
        "IDMODOPERACNFCE" => "T",
        "IDSITUVENDA"     => "T",
        "NRINSCRCONS"     => "T",
        "NRNOTAFISCALCE"  => "T"
    );
    
    public static $logMessages = array();
    
    public function __construct (Utilities\Util $utilService, TransmissaoNfceService $transmissaoService, NotaFiscal $notaFiscal, $instanceManager){
        $this->transmissaoService = $transmissaoService;
        $this->utilService = $utilService;
        $this->notaFiscal = $notaFiscal;
        $this->instanceManager = $instanceManager;
    }
    
    public static function escreverLog ($data) {
        array_push(self::$logMessages, $data);
    }
    
    public static function escreverLogArray($array) {
        foreach ($array as $linha) {
            self::escreverLog($linha);
        }
    }
    
    public static function getLog () {
        return self::$logMessages;
    }
    
    public function buscar (Request\Filter $request, Response $response) {
        $parametros = $this->buildParameters($request);
        ini_set('memory_limit', '1G');
        
        if($parametros[':NRNOTAFISCALCE'] != 'T'){
            $parametros[':NRNOTAFISCALCE'] = str_pad($parametros[':NRNOTAFISCALCE'], 9, '0', STR_PAD_LEFT);
        }
        
        if(!isset($parametros[':ALLFIEDLS'])){
            $parametros[':ALLFIEDLS'] = 'T';
        }
        $result = $this->transmissaoService->executarQuery(TransmissaoNfceService::BUSCAR_NOTAS, $parametros);
        if($result->error()) {
            throw new \Exception("Erro! ".$result->getMessage(), 1);
        } else {
            $response->addDataSet(new DataSet('vnd20000_transmissaonfce', $result->getResult()));
        }
    }
    
    public function buscarNumeracao (Request\Filter $request, Response $response) {
        $parametros = $this->buildParameters($request);
        if(!isset($parametros[':NRNOTAFISCALCE'])){
            $parametros[':NRNOTAFISCALCE'] = 'T';
        }
        $result = $this->transmissaoService->executarQuery(TransmissaoNfceService::ADQNUMERACAO, $parametros);
        $response->addDataSet(new DataSet('vnd20000_transmissaonfce_adqnumeracao', $result->getResult()));
    }
    
    public function buildParameters (Request\Filter $request) {
        $conditions = $request->getFilterCriteria()->getConditions();
        $params = array();
        
        foreach ($conditions as $filtro) {
            if (is_array($filtro['value'])) {
                // Especifico desse controller
                if ($filtro['columnName'] == "DTVENDA") {
                    if (!empty($filtro['value'][0]) && !empty($filtro['value'][1])) {
                        $params[':DATAINI'] = $filtro['value'][0];
                        $params[':DATAFIN'] = $filtro['value'][1];
                    }
                }
                if(strpos($filtro['columnName'], '|')){
                    $colunas = explode('|', $filtro['columnName']);
                    foreach($colunas as $k => $con){
                        $params[':ALLFIEDLS'] = $filtro['value'][0]['_ALL'];
                    }
                }else{
                  $params[':'.$filtro['columnName']] = implode("_", $filtro['value']);     
                }
            } else {
                $params[':'.$filtro['columnName']] = $filtro['value']; 
            }
        }
        // Especifico desse controller
        $params = $this->verificarTiposVendas($params);
        $params = $this->verificarTipoOperacao($params);
        $params = $this->verificarDatas($params);
        $params = $this->verificarCamposDefault($params);
        return $params;
    }
    
    public function verificarTiposVendas($parametros) {
        
        $parametros[':VND'] = empty($parametros[':VND']) ? 'T' : $parametros[':VND'];
        
        if ($parametros[':VND'] == 'S') {
            $parametros[':IDSITUVENDA'] = 'O';
        } else if ($parametros[':VND'] == 'N') {
            $parametros[':IDSITUVENDA'] = 'C';
        } else if ($parametros[':VND'] == 'T') {
            $parametros[':IDSITUVENDA'] = 'T';
        } 
        return $parametros;
    }
    
    public function verificarDatas($parametros) {
        
        if (empty($parametros[':DATAINI']) || empty($parametros[':DATAFIN'])) {
            $parametros[':DATAINI'] = 
            $parametros[':DATAFIN'] = date('d/m/Y');
        }
        
        return $parametros;
    }
    
    public function verificarTipoOperacao($parametros) {
        
        $parametros[':TIPO_EMISSAO'] = empty($parametros[':TIPO_EMISSAO']) ? 'T' : $parametros[':TIPO_EMISSAO'];
        if ($parametros[':TIPO_EMISSAO'] == 'N') {
            $parametros[':IDMODOPERACNFCE'] = '1';
        } else if ($parametros[':TIPO_EMISSAO'] == 'C') {
           $parametros[':IDMODOPERACNFCE'] = '9';
        } else if ($parametros[':TIPO_EMISSAO'] == 'T') {
            $parametros[':IDMODOPERACNFCE'] = 'T';
        } 
        return $parametros;
    }
    
    public function verificarCamposDefault($parametros) {
        $keys = array_keys($this->defaultFields);
        
        foreach ($keys as $key) {
            if (empty($parametros[':'.$key])) {
                $parametros[':'.$key] = $this->defaultFields[$key];
            }
        }
        
        return $parametros;
        
    }
    
    public function inutilizacaoNfe(Request\Row $request, Response $response) {
        $this->entityManager = $this->transmissaoService->entityManager;
        $row = $request->getRow();
        
        self::escreverLog("Log Inutilização NFC-e");
        self::escreverLog("Data da operação: ".date("d/m/Y H:i:s"));
        self::escreverLog("Data selecionada: ".$row['DTVENDA']['start'].' até '.$row['DTVENDA']['end']);
        self::escreverLog("Operador: ".$row['CDOPERADOR']);
        self::escreverLog("");
        
        $retorno = new Resposta();
        
        try {
            $this->entityManager->beginTransaction();
            $numerosFront = $row['NRNOTAFISCALCE'];
            $rows = $this->transmissaoService->getNotasInvalidacao($row);
            
            $numeros = array_filter(
                    $rows,
                    function ($e) use ($numerosFront) {
                        return in_array($e['NRNOTAFISCALCE'], $numerosFront);
                    }
                );
                
            $inutilizacao = new TransmissaoNFCE($this->transmissaoService, $numeros, $this->notaFiscal, $this->instanceManager, $this->utilService);
            $inutilizacao->inutilizarNotas($row);
            
            self::escreverLog("");
            self::escreverLog("## Confirmando alterações ##");
            $this->entityManager->getConnection()->commit();
            // $this->entityManager->rollback();
            self::escreverLog("Operação finalizada com sucesso! Data: ".date("d/m/Y H:i:s"));
            $retorno->setStatus(Resposta::SUCESSO);
            $retorno->setMessage(
                array(
                    'log'       => $this->getLog()
                )    
            );
        } catch (\Exception $e){
            self::escreverLog("");
            self::escreverLog(str_pad(' ERRO! ', strlen($e->getMessage()), "-", STR_PAD_BOTH));
            self::escreverLog("| ".$e->getMessage()." |");
            self::escreverLog(str_pad('-', strlen($e->getMessage()), "-", STR_PAD_LEFT));
            self::escreverLog("");
            self::escreverLog("## Desfazendo alterações ##");
            $this->entityManager->rollback();
            self::escreverLog("Operação finalizada de maneira inesperada. Data: ".date("d/m/Y H:i:s"));
            // $retorno->setMessage('Erro! <br><br>'.$e->getMessage().' Arquivo: '. $e->getFile() .' Linha: '.$e->getLine().'. <br><br>Todas as alterações foram revertidas.');
            $retorno->setMessage(
                array(
                    'text' => 'Erro! <br><br>'.$e->getMessage().'<br><br>Todas as alterações foram revertidas.',
                    'log'  =>  $this->getLog()
                )
            );
            $retorno->setStatus(Resposta::ERRO);
            $retorno->setError(true);
        }
        
        $response->addDataSet(new DataSet('resultado', $retorno->toPrint()));
    }
    
    public function transmitirnfecf(Request\Row $request, Response $response) {
        
        $this->entityManager = $this->transmissaoService->entityManager;
        $row = $request->getRow();
        
        self::escreverLog("Log Transmissão NFC-e");
        self::escreverLog("Data da operação: ".date("d/m/Y H:i:s"));
        self::escreverLog("Data selecionada: ".$row['DTVENDA']['start'].' até '.$row['DTVENDA']['end']);
        self::escreverLog("Operador: ".$row['CDOPERADOR']);
        self::escreverLog("");
        
        $retorno = new Resposta();
        try {
            
            $this->entityManager->beginTransaction();
            $transmissao = $this->getNotas($request);
            
            $this->escreverLog("Notas válidas: ".$transmissao->getQtdNotasValidas());
            $this->escreverLog(str_pad('-', 100, "-", STR_PAD_LEFT));
            
            $transmissao->verificarSituacaoNFCeSefaz();
            $transmissao->transmitirNotas();
            
            self::escreverLog("");
            self::escreverLog("## Confirmando alterações ##");
            $this->entityManager->getConnection()->commit();
            self::escreverLog("Operação finalizada com sucesso! Data: ".date("d/m/Y H:i:s"));
            $retorno->setStatus(Resposta::SUCESSO);
            $retorno->setMessage(
                array(
                    'notasValidas' => $transmissao->getQtdNotasValidas(),
                    'log'       => $this->getLog()
                )    
            );
            
        } catch (\Exception $e){
            self::escreverLog("");
            self::escreverLog(str_pad(' ERRO! ', strlen($e->getMessage()), "-", STR_PAD_BOTH));
            self::escreverLog("| ".$e->getMessage()." |");
            self::escreverLog(str_pad('-', strlen($e->getMessage()), "-", STR_PAD_LEFT));
            self::escreverLog("");
            self::escreverLog("## Desfazendo alterações ##");
            $this->entityManager->rollback();
            self::escreverLog("Operação finalizada de maneira inesperada. Data: ".date("d/m/Y H:i:s"));
            $retorno->setMessage(
                array(
                    'text' => 'Erro! <br><br>'.$e->getMessage().'<br><br>Todas as alterações foram revertidas.',
                    'log'  =>  $this->getLog()
                )
            );
            $retorno->setStatus(Resposta::ERRO);
            $retorno->setError(true);
        }
        $response->addDataSet(new DataSet('resultado', $retorno->toPrint()));
    }
    
    public function corrigirDuplicidade(Request\Row $request, Response $response) {
        
        $retorno = new Resposta();
        $retorno->setStatus(Resposta::ERRO);
        
        $row = $request->getRow();
        
        self::escreverLog("Log Correção de Duplicidade NFC-e");
        self::escreverLog("Data da operação: ".date("d/m/Y H:i:s"));
        self::escreverLog("Data selecionada: ".$row['DATAINI'].' até '.$row['DATAFIN']);
        self::escreverLog("Operador: ".$row['CDOPERADOR']);
        self::escreverLog("");
        
        $this->entityManager = $this->transmissaoService->entityManager;
        
        try {
            $this->entityManager->beginTransaction();
            
            $row = get_object_vars($request->getRow());
            $retorno->setParameters($row);
            $duplicidades = $this->transmissaoService->getDuplicidades($row);
            if (!empty($duplicidades)) {
                $correcao = new TransmissaoNFCE($this->transmissaoService, $row['rows'], $this->notaFiscal, $this->instanceManager, $this->utilService);
                $correcao->atualizaNFCE($duplicidades);
            } else {
                self::escreverLog("Não foi encontrada nenhuma nota em duplicidade.");
                self::escreverLog("");
                $retorno->setStatus(Resposta::NAO_COMPLETADO);
                $retorno->setMessage('Operação Realizada!<br>Não foi encontrada nenhuma nota enviada em duplicidade para a SEFAZ.');
            }

            self::escreverLog("## Confirmando alterações ##");
            $this->entityManager->getConnection()->commit();
            self::escreverLog("Operação finalizada com sucesso! Data: ".date("d/m/Y H:i:s"));
            // $this->entityManager->rollback();
            $retorno->setStatus(Resposta::SUCESSO);
            $retorno->setMessage(
                array(
                    'log'       => $this->getLog()
                )    
            );
        } catch (\Exception $e){
            self::escreverLog("");
            self::escreverLog(str_pad(' ERRO! ', strlen($e->getMessage()), "-", STR_PAD_BOTH));
            self::escreverLog("| ".$e->getMessage()." |");
            self::escreverLog(str_pad('-', strlen($e->getMessage()), "-", STR_PAD_LEFT));
            self::escreverLog("");
            self::escreverLog("## Desfazendo alterações ##");
            $this->entityManager->rollback();
            self::escreverLog("Operação finalizada de maneira inesperada. Data: ".date("d/m/Y H:i:s"));
            $retorno->setStatus(Resposta::ERRO);
            $retorno->setError(true);
            $retorno->setMessage(
                array(
                    'text' => 'Erro! <br><br>'.$e->getMessage().'<br><br>Todas as alterações foram revertidas.',
                    'log'  =>  $this->getLog()
                )
            );
        }
        $response->addDataSet(new DataSet('resultado', $retorno->toPrint()));
    }
    
    public function iniciarTentativas(Request\DataSet $request, Response $response) {
        $rows = $request->getDataSet()->getRows();
        foreach($rows as &$row) {
            if($row['IDSTATUSNFCE'] == 'P' || $row['IDSTATUSNFCE'] == 'R'){
                $updatadas = $this->transmissaoService->iniciarTentativas($row['CDFILIAL'], $row['CDCAIXA'], $row['NRSEQVENDA'], $row['NRORG']);
            }
        }
        $response->addNotification(new Notification("Número de tentativa reiniciado com sucesso.", Notification::TYPE_SUCCESS));
    }
    
    public function getNotas(Request\Row $request) {
        $row = get_object_vars($request->getRow());
        if (empty($row)) {
              throw new \Exception("Nenhuma nota selecionada. Função: getNotas.", 1);
        } else {
            $transmissao = new TransmissaoNFCE($this->transmissaoService, $row['rows'], $this->notaFiscal, $this->instanceManager, $this->utilService);
        }
        return $transmissao;
    }
    
    public function saveEdit(Request\DataSet $request, Response $response) {
        $parametros = array();
        $rows = $this->utilService->getParamsAsArray($request);
        $row = $rows[0];
        $NRINSCRCONS = $row['NRINSCRCONS'];
        
        if(!empty($NRINSCRCONS)){
            $NRINSCRCONS = preg_replace("/[^0-9]/", "", $NRINSCRCONS);
        	$NRINSCRCONS = str_pad($NRINSCRCONS, 11, '0', STR_PAD_LEFT);
        }
        
        if (strlen($NRINSCRCONS) < 12 && !empty($NRINSCRCONS)) {
            if (!$this->utilService->validarCPF($NRINSCRCONS)) {
                $response->addMessage(new Message("O CPF inserido não é válido.", Message::TYPE_WARNING)); 
                throw new \Exception("Erro ao validar o CPF/CNPJ");
            }
        } else if(!empty($NRINSCRCONS)){
            if (!$this->utilService->validarCNPJ($NRINSCRCONS)) {
                $response->addMessage(new Message("O CNPJ inserido não é válido.", Message::TYPE_WARNING)); 
                throw new \Exception("Erro ao validar o CPF/CNPJ");
            }  
        }
        
        $parametros[':CDFILIAL'] = $row['CDFILIAL'];
        $parametros[':CDCAIXA'] = $row['CDCAIXA'];
        $parametros[':NRSEQVENDA'] = $row['NRSEQVENDA'];
        $parametros[':NRNOTAFISCALCE'] = $row['NRNOTAFISCALCE'];
        $parametros[':NRINSCRCONS'] = !empty($NRINSCRCONS) ? $NRINSCRCONS : null;
        
        $this->transmissaoService->runQuery(TransmissaoNfceService::UPDATENRINSCRCONS, $parametros);
        
    }
    
    public function cancelarTransmissao(Request\Row $request, Response $response) {
        $retorno = new Resposta();
        $retorno->setStatus(Resposta::ERRO);
        $row = $request->getRow();
        
        self::escreverLog("Log Cancelar Transmissão NFC-e");
        self::escreverLog("Data da operação: ".date("d/m/Y H:i:s"));
        self::escreverLog("Operador: ".$row['CDOPERADOR']);
        self::escreverLog("");
        
        $this->entityManager = $this->transmissaoService->entityManager;
        
        try {
            $this->entityManager->beginTransaction();
            
            $cdsParaVend = $this->getParametrosVendas($row['CDFILIAL'], $row['NRORG']);

            $data1 = \DateTime::createFromFormat ('d/m/Y H:i:s', $cdsParaVend[0]['DTATUAL']);
            $data2 = \DateTime::createFromFormat ('d/m/Y H:i:s', empty($row['DTHRPROTOCONFCE']) ? $row['DTEMISSAONFCE'] : $row['DTHRPROTOCONFCE']);
            $data2->add(new \DateInterval('P1D'));
            
            if ($data1 > $data2) {
                   throw new \Exception("Operação Bloqueada! Não é permitido o cancelamento extemporâneo de Nota Fiscal Eletrônica.");
            } else {
                self::escreverLog(" "); 
                self::escreverLog("Cancelando Nota Fiscal...");  
                
                $result = $this->notaFiscal->cancelarNota($row);  
                
                $CDFILIAL = $row['CDFILIAL'];
                $CDCAIXA = $row['CDCAIXA'];
                $NRSEQVENDA = $row['NRSEQVENDA'];
                $DSRAZAOCANCNFCE = $row['JUSTIFICATIVA'];
                $DSARQXMLCANCNFCE = $result['xml'];
                $CDOPERADOR = $row['CDOPERADOR'];
                
                if(!empty($DSARQXMLCANCNFCE)){
                    $save = $this->notaFiscal->updateVendaStatusNfceCan($CDFILIAL, $CDCAIXA, $NRSEQVENDA, $DSRAZAOCANCNFCE, $DSARQXMLCANCNFCE, $CDOPERADOR);
                
                    if ($save['error']) {
                        throw new \Exception("Ocorreu um erro ao salvar as informações de cancelamento. ".$save['message']);
                    }
                }else{
                    throw new \Exception("Não foi possível realizar o cancelamento da Nota, informações faltantes!");
                }
                self::escreverLog(" "); 
            }

            self::escreverLog("## Confirmando alterações ##");
            $this->entityManager->getConnection()->commit();
            self::escreverLog("Operação finalizada com sucesso! Data: ".date("d/m/Y H:i:s"));
            $retorno->setStatus(Resposta::SUCESSO);
            $retorno->setMessage(
                array(
                    'log'             => $this->getLog(),
                    'DSOBSSTATUSNFCE' => isset($save['DSOBSSTATUSNFCE']) ? $save['DSOBSSTATUSNFCE'] : ''
                )    
            );
        } catch (\Exception $e){
            self::escreverLog("");
            self::escreverLog(str_pad(' ERRO! ', strlen($e->getMessage()), "-", STR_PAD_BOTH));
            self::escreverLog("| ".$e->getMessage()." |");
            self::escreverLog(str_pad('-', strlen($e->getMessage()), "-", STR_PAD_LEFT));
            self::escreverLog("");
            self::escreverLog("## Desfazendo alterações ##");
            $this->entityManager->rollback();
            self::escreverLog("Operação finalizada de maneira inesperada. Data: ".date("d/m/Y H:i:s"));
            $retorno->setStatus(Resposta::ERRO);
            $retorno->setError(true);
            $retorno->setMessage(
                array(
                    'text' => 'Erro! <br><br>'.$e->getMessage().'<br><br>Todas as alterações foram revertidas.',
                    'log'  =>  $this->getLog()
                )
            );
        }
        
        $response->addDataSet(new DataSet('resultado', $retorno->toPrint()));
    }
    
    public function cancelamentoSbTransmissao(Request\Row $request, Response $response){
        $retorno = new Resposta();
        $retorno->setStatus(Resposta::ERRO);
        $row = $request->getRow();
        
        
        self::escreverLog("Log Cancelar Transmissão NFC-e");
        self::escreverLog("Data da operação: ".date("d/m/Y H:i:s"));
        self::escreverLog("Operador: ".$row['CDOPERADOR']);
        self::escreverLog("");
        
        
        $this->entityManager = $this->transmissaoService->entityManager;
        
        try {
            $this->entityManager->beginTransaction();
            
            self::escreverLog(" "); 
            self::escreverLog("Cancelando Nota Fiscal...");  
            $result = $this->notaFiscal->cancelarNotaSubstituicao($row);  
            
            if(!empty($result)){
                
                if ($result['error']) {
                    self::escreverLog("Operação finalizada de maneira inesperada. Data: ".date("d/m/Y H:i:s"));
                    throw new \Exception("Ocorreu um erro ao salvar as informações de cancelamento. ".$result['message']);
                }else{
                    
                    $CDFILIAL = $row['CDFILIAL'];
                    $CDCAIXA = $row['CDCAIXA'];
                    $NRSEQVENDA = $row['NRSEQVENDA'];
                    $DSRAZAOCANCNFCE = $row['JUSTIFICATIVA'];
                    $DSARQXMLCANCNFCE = $result['xml'];
                    $CDOPERADOR = $row['CDOPERADOR'];
                    $DSOBSSTATUSNFCE = $row['DSOBSSTATUSNFCE'];
                    
                    $save = $this->notaFiscal->updateVendaStatusNfceCanSB($CDFILIAL, $CDCAIXA, $NRSEQVENDA, $DSRAZAOCANCNFCE, $DSARQXMLCANCNFCE, $CDOPERADOR, $DSOBSSTATUSNFCE);
                    
                    if ($save['error']) {
                        throw new \Exception("Ocorreu um erro ao salvar as informações de cancelamento. ".$save['message']);
                    }
                    
                    self::escreverLog(" "); 
                    self::escreverLog("## Confirmando alterações ##");
                    $this->entityManager->getConnection()->commit();
                    self::escreverLog("Operação finalizada com sucesso! Data: ".date("d/m/Y H:i:s"));
                    $retorno->setStatus(Resposta::SUCESSO);
                    $retorno->setMessage(
                        array(
                            'log'             => $this->getLog(),
                            'DSOBSSTATUSNFCE' => isset($save['DSOBSSTATUSNFCE']) ? $save['DSOBSSTATUSNFCE'] : ''
                        )    
                    );
                }
            }else{
                self::escreverLog("Operação finalizada de maneira inesperada. Data: ".date("d/m/Y H:i:s"));
                throw new \Exception("Ocorreu um erro ao realizar cancelamento");
            }
        
        } catch (\Exception $e){
            self::escreverLog("");
            self::escreverLog(str_pad(' ERRO! ', strlen($e->getMessage()), "-", STR_PAD_BOTH));
            self::escreverLog("| ".$e->getMessage()." |");
            self::escreverLog(str_pad('-', strlen($e->getMessage()), "-", STR_PAD_LEFT));
            self::escreverLog("");
            self::escreverLog("## Desfazendo alterações ##");
            $this->entityManager->rollback();
            self::escreverLog("Operação finalizada de maneira inesperada. Data: ".date("d/m/Y H:i:s"));
            $retorno->setStatus(Resposta::ERRO);
            $retorno->setError(true);
            $retorno->setMessage(
                array(
                    'text' => 'Erro! <br><br>'.$e->getMessage().'<br><br>Todas as alterações foram revertidas.',
                    'log'  =>  $this->getLog()
                )
            );
        }
        $response->addDataSet(new DataSet('resultado', $retorno->toPrint()));
    }
    
    private function getParametrosVendas($CDFILIAL, $NRORG) {
        $parametros = array(
            ':CDFILIAL' => $CDFILIAL,
            ':NRORG'    => $NRORG
        );
        $result = $this->transmissaoService->executarQuery(TransmissaoNfceService::ADQPARAVEND, $parametros);
        
        if ($result->error())
            throw new \Exception("Erro! ".$result->getMessage(), 1);
        
        $retorno = $result->getResult();
        return $retorno;
    }
    
}