<?php

namespace Teknisa\MAN\Services;

use Doctrine\DBAL\Connection;
use Teknisa\MAN\Utilities;

class FechamentoCaixaService {

    protected $conn;

    public function __construct(Connection $conn) {
        $this->conn = $conn;
    }

    public function findCodigo($cdmodelo, $nrorg) {
        $params = array(
            "CDMODELO" => $cdmodelo,
            "NRORG" => $nrorg
     	); 
        return $this->conn->fetchAll(self::GET_CDMODELO,$params);
    }

    public function isReducaoZApproved($CDFILIAL,$CDCAIXA,$DTABERCAIX,$NRORG){
        
        $params = array(
            ':CDFILIAL'     => $CDFILIAL,
            ':CDCAIXA'      => $CDCAIXA,
            ':DTABERCAIX'   => $DTABERCAIX,
            ':NRORG'        => $NRORG
        );
        
        $reducaoZ = $this->conn->fetchAll(self::IS_REDUCAOZ_APPROVED,$params);
        
        if(empty($reducaoZ))
            $reducaoZ = false ;
        else
            $reducaoZ = true ;
        
        return $reducaoZ ; 
    }
    
    public function isNFCeApproved($CDFILIAL,$CDCAIXA,$DTABERCAIX,$NRORG){
        
        $params = array(
            ':CDFILIAL'     => $CDFILIAL,
            ':CDCAIXA'      => $CDCAIXA,
            ':DTABERCAIX'   => $DTABERCAIX,
            ':NRORG'        => $NRORG
        );
        
        $isNFCeApproved = $this->conn->fetchAll(self::IS_VENDA_NFCE_APPROVED,$params);
        if(empty($isNFCeApproved))
            $isNFCeApproved = false ;
        else
            $isNFCeApproved = true ;
        
        return $isNFCeApproved ; 
    }
    
    public function vendaPendenteExists($CDFILIAL,$CDCAIXA,$DTABERCAIX,$NRORG){
        
        $params = array(
            ':CDFILIAL'     => $CDFILIAL,
            ':CDCAIXA'      => $CDCAIXA,
            ':DTABERCAIX'   => $DTABERCAIX,
            ':NRORG'        => $NRORG
        );
        $vendaPendenteExists = $this->conn->fetchAll(self::VENDA_PENDENTE_EXISTS,$params);
        if(empty($vendaPendenteExists))
            $vendaPendenteExists = false ;
        else
            $vendaPendenteExists = true ;
        
        return $vendaPendenteExists ; 
    }
    
    public function fecharCaixa($CDFILIAL,$CDCAIXA,$DTABERCAIX,$DTFECHCAIX,$CDOPERFECH,$CDOPERADOR,$NRORG,$DTULTATU,$NRORGULTATU,$CDOPERULTATU){
        
        $params = array(
            ':CDFILIAL'     =>  $CDFILIAL,
            ':CDCAIXA'      =>  $CDCAIXA,
            ':DTABERCAIX'   =>  $DTABERCAIX,
            ':DTFECHCAIX'   =>  $DTFECHCAIX,
            ':CDOPERFECH'   =>  $CDOPERFECH,
            ':CDOPERADOR'   =>  $CDOPERADOR,
            ':NRORG'        =>  $NRORG,
            ':DTULTATU'     =>  $DTULTATU,
            ':NRORGULTATU'  =>  $NRORGULTATU,
            ':CDOPERULTATU' =>  $CDOPERULTATU
        );
        
        return $this->conn->executeQuery(self::FECHAR_CAIXA, $params);
    }
    
    public function deleteFechamento($CDFILIAL,$CDCAIXA,$DTABERCAIX,$CDOPERADOR,$NRORG,$DTULTATU,$NRORGULTATU,$CDOPERULTATU, $DTFECHCAIXCP){
        $params = array(
            ':CDFILIAL' =>$CDFILIAL,
            ':CDCAIXA'  =>$CDCAIXA,
            ':DTABERCAIX'   =>$DTABERCAIX,
            ':CDOPERADOR'   =>$CDOPERADOR,
            ':NRORG'    =>$NRORG,
            ':DTULTATU' =>$DTULTATU,
            ':NRORGULTATU'  =>$NRORGULTATU,
            ':DTFECHCAIXCP' =>$DTFECHCAIXCP,
            ':CDOPERULTATU' =>$CDOPERULTATU
        );
        
        return $this->conn->executeQuery(self::DELETE_FECHAMENTO_CAIXA, $params);
    }
    
    const GET_CDMODELO = "
    SELECT CDMODELO 
      FROM MODELO 
      WHERE CDMODELO = :CDMODELO 
        AND NRORG = :NRORG
    ";
    
  // Query para validar se a redução Z do caixa selecionado já foi aprovada.
  
    const IS_REDUCAOZ_APPROVED = "
    SELECT RZ.CDFILIAL, RZ.CDCAIXA, RZ.DTREDUCAOZ, 
           RZ.NRECFREDZ, RZ.IDSTATUS 
    FROM REDUCAOZ RZ
    WHERE CDFILIAL   = :CDFILIAL AND
          CDCAIXA    = :CDCAIXA  AND
          (NRORG       = :NRORG  ) AND
          DTREDUCAOZ = :DTABERCAIX AND
          ((IDSTATUS)   = 'E' OR (IDSTATUS = 'A'))";
          
  
  //Se o caixa não for ECF (IDTPEMISSAOFOS)
  //Se não for vazio, "Operação não permitida. As vendas NFC-e ou CF-e-SAT já foram aprovadas."
  
    const IS_VENDA_NFCE_APPROVED = "
    SELECT IDAPROVACAONFCE
    FROM
       VENDA
    WHERE
       (CDFILIAL  = :CDFILIAL  ) AND
       (CDCAIXA   = :CDCAIXA   ) AND
       (NRORG     = :NRORG  ) AND
       (DTABERTUR = :DTABERCAIX) AND
       (IDAPROVACAONFCE = 'A'    )";
  
  
  // Se o caixa tiver venda pendente, não deixa fechar.
  // "Operação não permitida. Este Caixa possui venda pendente."

    const VENDA_PENDENTE_EXISTS = "
    SELECT NRSEQVENDA
    FROM
       VENDA
    WHERE
       (CDFILIAL    = :CDFILIAL ) AND
       (CDCAIXA     = :CDCAIXA  ) AND
       (NRORG       = :NRORG  ) AND
       (IDSITUVENDA = 'P'         ) AND
       (DTABERTUR   = TO_DATE(:DTABERCAIX, 'DD/MM/YYYY HH24:MI:SS'))";
  
    const FECHAR_CAIXA = "
    UPDATE TURCAIXA
    SET 
      CDOPERFECH    = :CDOPERFECH,
      DTFECHCAIX    = :DTFECHCAIX,
      DTULTATU      = :DTULTATU,
      NRORGULTATU   = :NRORGULTATU,
      CDOPERULTATU  = :CDOPERULTATU
    WHERE CDFILIAL   = :CDFILIAL
      AND CDCAIXA    = :CDCAIXA
      AND DTABERCAIX = :DTABERCAIX
      AND NRORG      = :NRORG
      AND CDOPERADOR = :CDOPERADOR

    ";
  
    const DELETE_FECHAMENTO_CAIXA = "
    UPDATE TURCAIXA
    SET 
      CDOPERFECH    = NULL,
      DTFECHCAIX    = NULL,
      DTULTATU      = :DTULTATU,
      NRORGULTATU   = :NRORGULTATU,
      CDOPERULTATU  = :CDOPERULTATU,
      DTFECHCAIXCP  = :DTFECHCAIXCP
      
    WHERE CDFILIAL   = :CDFILIAL
      AND CDCAIXA    = :CDCAIXA
      AND DTABERCAIX = :DTABERCAIX
      AND NRORG      = :NRORG
      AND CDOPERADOR = :CDOPERADOR
    ";
}